/*
	$Header: /home/itig/cvsroot/ext2ifs/src/block.c,v 1.32 2004/01/23 09:28:13 jn Exp $
*/

#include "ext2fs.h"

/*
	This is all the ext2 block related code.

	Data types used in this file:

    Name		Type		Constraints
	BlockNo		__u32		This is a 1 based number.  0 means no block
	BlockCount	__u32		This is a 0 based count (ie 1 = 1 block)
				BLOCK_GROUP
	INodeNumber	__u32		This is a 1 based number.  0 means no INode
	GroupNumber __u32		This is a 1 based number
	INodeInGroup __u32		This is 0 based.  It is used in conjunction with a GroupNumber

*/


NTSTATUS
CompletionRoutine (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID Context
    )
{
	KEVENT *Event = (KEVENT *)Context;

	/*KdPrint(("[EXT2]: In CompletionRoutine status = 0x%08x\n", Irp->IoStatus.Status));
	KdPrint(("[EXT2]: In CompletionRoutine information = 0x%08x\n", Irp->IoStatus.Information));*/

    //
    //  Store the correct information field into the Irp.
    //

/*    if (!NT_SUCCESS( Irp->IoStatus.Status )) {

        Irp->IoStatus.Information = 0;
    }*/
//    Irp->IoStatus.Information = 0;

    //KeSetEvent( Event, 0, FALSE );

    return Irp->IoStatus.Status;//MORE_PROCESSING_REQUIRED;
}


/*
 This is used to check the device at mount time
*/

NTSTATUS
XtBLReadBlockDevice (
	IN PDEVICE_OBJECT	DeviceObject,
	IN PLARGE_INTEGER	Offset,
	IN ULONG			Length,
	OUT PVOID			Buffer
	)
{
	KEVENT			event;
	PIRP			irp;
	IO_STATUS_BLOCK io_status;
	NTSTATUS		status;

	//PAGED_CODE();
	ASSERT(DeviceObject != NULL);
	ASSERT(Offset != NULL);
	ASSERT(Buffer != NULL);
	
	//KdPrint(("[EXT2]: XtBLReadBlockDevice length 0x%08x offset 0x%08x\n", Length, Offset->QuadPart));

	KeInitializeEvent(&event, NotificationEvent, FALSE);

	irp = IoBuildSynchronousFsdRequest(
		IRP_MJ_READ,
		DeviceObject,
		Buffer,
		Length,
		Offset,
		&event,
		&io_status
		);

	if (!irp)
	{
		return STATUS_INSUFFICIENT_RESOURCES;
	}

    SetFlag( IoGetNextIrpStackLocation( irp )->Flags, SL_OVERRIDE_VERIFY_VOLUME );


	status = IoCallDriver(DeviceObject, irp);

	if (status == STATUS_PENDING)
	{
		KeWaitForSingleObject(
			&event,
			Suspended,
			KernelMode,
			FALSE,
			NULL
			);
		status = io_status.Status;
	}

	return status;
}


/*
  This is used to do real read requests

  I am not sure how or why this works but it appears to be reliable for hard disks
  It does not seem to work for floppy disks...
*/

NTSTATUS
XtBLReadBlockDeviceSync (
    IN PFSD_IRP_CONTEXT IrpContext,
	IN PtrXtFSVCB		Vcb,
    IN PLARGE_INTEGER   Offset,
    IN ULONG            Length,
    IN OUT PVOID        Buffer
    )
{
    KEVENT				Event;
    IO_STATUS_BLOCK		IoStatus;
    NTSTATUS			Status;
    PIO_STACK_LOCATION	IoStack;
    PIO_STACK_LOCATION	IrpSp;

    PAGED_CODE();

	ASSERT(IrpContext);
	ASSERT(Vcb);
    ASSERT(Offset);
    ASSERT(Buffer);

//	KdPrint(("[EXT2]: XtBLReadBlockDeviceSync length 0x%08x offset 0x%08x %08x\n", Length, Offset->HighPart, Offset->LowPart));

/*    KeInitializeEvent(&Event, NotificationEvent, FALSE);

    IoSetCompletionRoutine( IrpContext->Irp,
                            CompletionRoutine,
                            &Event,
                            TRUE,
                            TRUE,
                            TRUE );*/

    IrpSp = IoGetNextIrpStackLocation( IrpContext->Irp );

    //
    //  Setup the Stack location to do a read from the disk driver.
    //

    IrpSp->MajorFunction = IRP_MJ_READ;
    IrpSp->Parameters.Read.Length = Length;
    IrpSp->Parameters.Read.ByteOffset.QuadPart = Offset->QuadPart;

	SetFlag( IoGetNextIrpStackLocation( IrpContext->Irp )->Flags, SL_OVERRIDE_VERIFY_VOLUME );

	IrpContext->Completed = TRUE;
    Status = IoCallDriver( Vcb->TargetDeviceObject, IrpContext->Irp );

	/*
		I am not sure what goes on here, but I think that the disk will complete the 
		Irp, which will in turn return execution to the point at which the original
		Irp was dispatched.
	*/

/*	if(Status == STATUS_PENDING)
	{
		KeWaitForSingleObject(
			&Event,
			Executive,
			KernelMode,
			FALSE,
			NULL
			);
		Status = IrpContext->Irp->IoStatus.Status;
	}*/

	// wait for the completion to execute...
	//KdPrint(("[EXT2]: KeWaitForSingleObject\n"));
/*    KeWaitForSingleObject( &Event,
                           Executive,
                           KernelMode,
                           FALSE,
                           NULL );

	Status = IrpContext->Irp->IoStatus.Status;
	IrpContext->Irp->IoStatus.Information = 0;*/

//	KdPrint(("[EXT2]: Read done\n"));

    return Status;
}


/*
NTSTATUS
FsdReadBlockDevice (
    IN PDEVICE_OBJECT   DeviceObject,
    IN PLARGE_INTEGER   Offset,
    IN ULONG            Length,
    IN OUT PVOID        Buffer
    )
{
    KEVENT          Event;
    PIRP            Irp;
    IO_STATUS_BLOCK IoStatus;
    NTSTATUS        Status;

    PAGED_CODE();

    ASSERT(DeviceObject != NULL);
    ASSERT(Offset != NULL);
    ASSERT(Buffer != NULL);

    KeInitializeEvent(&Event, NotificationEvent, FALSE);

    Irp = IoBuildSynchronousFsdRequest(
        IRP_MJ_READ,
        DeviceObject,
        Buffer,
        Length,
        Offset,
        &Event,
        &IoStatus
        );

    if (!Irp)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    Status = IoCallDriver(DeviceObject, Irp);

    if (Status == STATUS_PENDING)
    {
        KeWaitForSingleObject(
            &Event,
            Suspended,
            KernelMode,
            FALSE,
            NULL
            );
        Status = IoStatus.Status;
    }

    return Status;
}
*/



/*
	These are debug support routines
*/
VOID
DumpBlock(PVOID Block,
		  PtrXtFSVCB Vcb,
		  int count)
{
	int b; // block
	int rows; // number of rows
	int columns; // number of columns
	int row; // current row
	int column; // current column
	unsigned char *Buffer = (unsigned char *) Block;

	columns = 32;
	rows = Vcb->BLBlockSize / columns;

	for(b=0; b<count; b++)
	{
		for(row = 0; row < rows; row++)
		{
			for(column = 0; column < columns; column++)
			{
				KdPrint(("%02x ", *Buffer++));
			}
			KdPrint(("\n"));
		}
		KdPrint((""));
	}
}

VOID
DumpDirBlock(PVOID Block,
		  PtrXtFSVCB Vcb)
{
	char *data = Block;
	struct ext2_dir_entry_2 *dir;
	unsigned int offset = 0;
	int i;
	char buffer[512];

	ASSERT(Block);
	ASSERT(Vcb);

	while(offset < Vcb->BLBlockSize)
	{
		dir = (struct ext2_dir_entry_2 *)(data + offset);

		if(dir->inode != 0)
		{
			KdPrint(("%6d ", dir->inode));
			for(i = 0; i< dir->name_len; i++)
			{
				if(i >= 512)
				{
					break;
				}
				buffer[i] = dir->name[i];
				//KdPrint(("%c", dir->name[i]));
			}
			buffer[i++] = 0;
			KdPrint(("%s\n", buffer));
			//KdPrint((" = %d (%d %d %d)\n", dir->inode, dir->name_len, dir->rec_len, dir->file_type));
		}

		// advance to next record
		offset += dir->rec_len;
		//KdPrint(("offset = %d\n", offset));

		if(dir->rec_len == 0)
		{
			break;
		}
	}
}


VOID
DumpINode(PINODE INode)
{
	int i;

	KdPrint(("[EXT2]: mode   %04X\n", INode->i_mode));
	KdPrint(("[EXT2]: uid    %04X\n", INode->i_uid));
	KdPrint(("[EXT2]: size   %08X\n", INode->i_size));
	KdPrint(("[EXT2]: atime  %08X\n", INode->i_atime));
	KdPrint(("[EXT2]: ctime  %08X\n", INode->i_ctime));
	KdPrint(("[EXT2]: mtime  %08X\n", INode->i_mtime));
	KdPrint(("[EXT2]: dtime  %08X\n", INode->i_dtime));
	KdPrint(("[EXT2]: gid    %04X\n", INode->i_gid));
	KdPrint(("[EXT2]: links  %04X\n", INode->i_links_count));
	KdPrint(("[EXT2]: blocks %08X\n", INode->i_blocks));
	KdPrint(("[EXT2]: flags  %08X\n", INode->i_flags));

	for(i=0; i<EXT2_N_BLOCKS; i++)
	{
		KdPrint(("[EXT2]: block[%02d]  %08X\n", i, INode->i_block[i]));
	}

}

VOID
DumpDirectoryINode(PINODE_OBJECT INode,
  		  PtrXtFSVCB Vcb
)
{
	PDIR_OBJECT dir;
	//PBLOCK_OBJECT block;
	ULONG i;
	WCHAR buffer[EXT2_NAME_LEN];
	__u32 inode;
	ULONG len;
	NTSTATUS Status;
	char str[512];

	ASSERT(INode);


	// make sure this is a directory...
	//if(INode->i_flags

/*	for(i = 0; i< EXT2_IND_BLOCK; i++)
	{
		if(INode->i_block[i])
		{
			XtLockBlock(Vcb, INode->i_block[i], 1, &block);

			DumpDirBlock(block->Buffer, Vcb);

			XtUnLockBlock(Vcb, &block);
		}
	}*/

	XtCreateDirObject(Vcb, INode, &dir, NULL);
	while(1)
	{
		len = EXT2_NAME_LEN * 2;

		Status = XtDirGetEntry(dir, buffer, &len, &inode);

		if(Status == STATUS_SUCCESS)
		{
			// print out the entry
			len = len / 2;
			for(i = 0; i<len; i++)
			{
				if(i >= 512)
				{
					break;
				}
				str[i] = (char)buffer[i];
				//KdPrint(("%c", buffer[i]));
			}
			str[i++] = 0;
			KdPrint(("%s %d\n", str, inode));
		}
		else
		{
			break;
		}
	}
	XtDestroyDirObject(&dir);
}



/*
	Reading a block, we need to know
	total number of blocks
	size of a block

	when a block is read, it is added to a list of allocated blocks
	when it is freed, it count is decremented.
*/

NTSTATUS 
XtBLInitBlockIO(
	PtrXtFSVCB Vcb
)
{
	NTSTATUS		Status = STATUS_SUCCESS;
	PINODE_OBJECT	RootINode;
    CC_FILE_SIZES	FileSizes;
	__u32			PointersPerBlock;

	KdPrint(("[EXT2]: XtBLInitBlockIO %d\n", KeGetCurrentIrql()));

	ASSERT(Vcb);
	ASSERT(Vcb->SuperBlock);

	ACQUIRE_EX(Vcb->Resource);

	switch(Vcb->SuperBlock->s_log_block_size)
	{
	case 0:
		Vcb->BLBlockSize = 1024;
		break;

	case 1:
		Vcb->BLBlockSize = 2048;
		break;

	case 2:
		Vcb->BLBlockSize = 4096;
		break;

	default:
		Vcb->BLBlockSize = 0;
		Status = STATUS_INVALID_PARAMETER;
		break;
	}

	KdPrint(("[EXT2]: Block size = %d\n", Vcb->BLBlockSize));
	KdPrint(("[EXT2]: Block count = %d\n", Vcb->SuperBlock->s_blocks_count));

	PointersPerBlock = Vcb->BLBlockSize / 4;
	Vcb->BLIndMax = PointersPerBlock + EXT2_NDIR_BLOCKS;
	Vcb->BLDIndMax = PointersPerBlock * PointersPerBlock + Vcb->BLIndMax;
	Vcb->BLTIndMax = PointersPerBlock * PointersPerBlock * PointersPerBlock + Vcb->BLDIndMax;

	KdPrint(("[EXT2]: MAX  Indirect = %d\n", Vcb->BLIndMax));
	KdPrint(("[EXT2]: MAX DIndirect = %d\n", Vcb->BLDIndMax));
	KdPrint(("[EXT2]: MAX TIndirect = %d\n", Vcb->BLTIndMax));

	/*Vcb->CommonFCBHeader.AllocationSize.QuadPart  = Vcb->BLBlockSize;
	Vcb->CommonFCBHeader.FileSize.QuadPart        = Vcb->BLBlockSize * Vcb->SuperBlock->s_blocks_count;
	Vcb->CommonFCBHeader.ValidDataLength.QuadPart = Vcb->PartitionInformation.PartitionLength.QuadPart;
//	Vcb->CommonFCBHeader.ValidDataLength         = LargeIntegerMax;

	FileSizes.AllocationSize  = Vcb->CommonFCBHeader.AllocationSize;
	FileSizes.FileSize        = Vcb->CommonFCBHeader.FileSize;
	FileSizes.ValidDataLength = Vcb->CommonFCBHeader.ValidDataLength;


	CcSetFileSizes(Vcb->VirtualVolumeFile, &FileSizes);*/

	// FileOffset->QuadPart + (LONGLONG)Length ) <= SharedCacheMap->SectionSize.QuadPart


	Vcb->BGGroupCount = XtDiv2(Vcb->SuperBlock->s_blocks_count, Vcb->SuperBlock->s_blocks_per_group);
	KdPrint(("[EXT2]: %d Block Groups of %d blocks\n", Vcb->BGGroupCount, Vcb->SuperBlock->s_blocks_per_group));

	Vcb->BGDescriptorSize = XtDiv2((Vcb->BGGroupCount * sizeof(GROUP_DESC)), Vcb->BLBlockSize);
	KdPrint(("[EXT2]: Blocks group descriptors need %d blocks\n", Vcb->BGDescriptorSize));

	

	Vcb->BGGroupDescriptors = XtBLGetBlocks(Vcb, Vcb->SuperBlock->s_first_data_block + 1, Vcb->BGDescriptorSize, &(Vcb->BGGroupDescriptorsBcb));
//    DumpBlock(Vcb->BGGroupDescriptors, Vcb->BLBlockSize, 1);

    // blocks in bitmaps? (8 = 8 bits per byte)
	Vcb->BGBlockMapSize = XtDiv2(XtDiv2(Vcb->SuperBlock->s_blocks_per_group, 8), Vcb->BLBlockSize);
    KdPrint(("[EXT2]: Block bitmap %d blocks\n", Vcb->BGBlockMapSize));

    // bocks in inode bitmap
    Vcb->BGINodeMapSize = XtDiv2(XtDiv2(Vcb->SuperBlock->s_inodes_per_group, 8), Vcb->BLBlockSize);
    KdPrint(("[EXT2]: INode bitmap %d blocks\n", Vcb->BGINodeMapSize));

	Vcb->BGTableSize = XtDiv2(Vcb->SuperBlock->s_inodes_per_group * sizeof(INODE), Vcb->BLBlockSize);
//	Vcb->BGTableSize = XtDiv2(Vcb->SuperBlock->s_inodes_per_group * Vcb->SuperBlock->s_inode_size, Vcb->BLBlockSize);

    KdPrint(("[EXT2]: inode table %d blocks\n", Vcb->BGTableSize));

	Vcb->BGBlockHeadSize = Vcb->BGBlockMapSize + Vcb->BGINodeMapSize + Vcb->BGTableSize;
	Vcb->BGBlockGroups = (PBLOCK_GROUP *)XtMMAllocatePool(NonPagedPool, sizeof(PBLOCK_GROUP) * Vcb->BGGroupCount);
	ASSERT(Vcb->BGBlockGroups);


	KdPrint(("[EXT2]: sizeof(INODE) = %d  s_inode_size = %d\n", sizeof(INODE), Vcb->SuperBlock->s_inode_size));
	Vcb->INodesPerBlock = Vcb->BLBlockSize / sizeof(INODE);// Vcb->SuperBlock->s_inode_size ??

	RELEASE(Vcb->Resource);

/*	KdPrint(("[EXT2]: Loading root INode\n"));

	// read in the root inode
	XtLockINode(Vcb, EXT2_ROOT_INO, &RootINode);
	DumpINode(RootINode->INode);
	DumpDirectoryINode(RootINode, Vcb);*/

	return Status;
}



/*
  This will map the requested blocks into memory and pin them.
  This should not be used directly.  Use BLOCK_OBJECT's insetead
*/
PVOID
XtBLGetBlocks(
	PtrXtFSVCB Vcb,
	__u32 BlockNo,
	__u32 BlockCount,
	PBCB *BlockBcb
)
{
	PVOID Buffer;
	ULONG Length;
	LARGE_INTEGER Offset;
	NTSTATUS Status;

	ASSERT(Vcb);

//	KdPrint(("[EXT2]: Reading %d blocks from = %d\n", BlockCount, BlockNo));

	Length = BlockCount * Vcb->BLBlockSize;
	Offset = RtlEnlargedIntegerMultiply(Vcb->BLBlockSize, BlockNo);

//	KdPrint(("[EXT2]: Reading %d bytes from = %I64d\n", Length, Offset.QuadPart));
//	KdPrint(("[EXT2]: About to map blocks\n"));

	if(CcPinRead(Vcb->VirtualVolumeFile, &Offset, Length, TRUE, BlockBcb, &Buffer))
	{
//		KdPrint(("[EXT2]: Map done\n"));
	}
	else
	{
		KdPrint(("[EXT2]: Map failed\n"));
		Buffer = 0;
	}

	return Buffer;
}

NTSTATUS
XtBLFreeBlocks(
	PtrXtFSVCB Vcb,
	PBCB *BlockBcb
)
{
	ASSERT(Vcb);
	ASSERT(BlockBcb);

	if(*BlockBcb)
	{
		//KdPrint(("[EXT2]: CcUnpinData being called IRQL = %d\n", KeGetCurrentIrql()));
		CcUnpinData(*BlockBcb);
	}
	return STATUS_SUCCESS;
}

/*
	This calculated Neumerator / Denominator
	but rounds up
*/
__u32 XtDiv2(
	__u32 Neumerator,
	__u32 Denominator
)
{
	__u32 Result = Neumerator / Denominator;
	if((Neumerator % Denominator) > 0)
	{
		Result++;
	}
	return Result;
}

/*
	This calculated Neumerator / Denominator
	and returns the whole part and the remainder
*/
__u32 XtDiv1(
	__u32 Neumerator,
	__u32 Denominator,
	__u32 *Remainder
)
{
	__u32 Result = Neumerator / Denominator;
	if(Remainder)
	{
		*Remainder = Neumerator % Denominator;
	}
	//KdPrint(("[EXT2]: XtDiv1 %u / %u -> %u r %u\n", Neumerator, Denominator, Result, *Remainder));

	return Result;
}

/*
	This calculated Neumerator / Denominator
	and returns the whole part and the remainder
*/
__u32 XtDiv64(
	ULONGLONG Neumerator,
	__u32 Denominator,
	__u32 *Remainder
)
{
	__u32 Result = (__u32)(Neumerator / Denominator);
	if(Remainder)
	{
		*Remainder = (__u32)(Neumerator % (ULONGLONG)Denominator);
	}
	//KdPrint(("[EXT2]: XtDiv64 %u / %u -> %u r %u\n", (__u32)Neumerator, Denominator, Result, *Remainder));

	return Result;
}

/*
	BlockGroup is freshly allocated memory
	we have to fill it in with the data off the disk
*/
NTSTATUS
XtBGLoadNewGroup(
	PtrXtFSVCB Vcb,
	PBLOCK_GROUP BlockGroup,
	__u32 GroupNumber
)
{
	int i;
	ASSERT(Vcb);

	//KdPrint(("[EXT2]: Loading in block group %d\n", GroupNumber));
	// read data from disk...
	BlockGroup->GroupNo = GroupNumber;
	BlockGroup->RefCount = 0;
	BlockGroup->DeviceObject = Vcb->TargetDeviceObject;

	// work out the location of our descriptor
	BlockGroup->Info = &(Vcb->BGGroupDescriptors[GroupNumber - 1]);

	ASSERT(BlockGroup->Info);

    BlockGroup->StartBlock = BlockGroup->Info->bg_block_bitmap;
//	KdPrint(("[EXT2]: Blocks group %d start at %0x\n", GroupNumber, BlockGroup->StartBlock));


	// load the block bitmap
#ifdef RWSUPPORT
	// We don't need the bitmaps until
	// we want to find a free one (for writing)
	BlockGroup->BlockBitmap = XtBLGetBlocks(Vcb, BlockGroup->Info->bg_block_bitmap, Vcb->BGBlockMapSize, &(BlockGroup->BlockBitmapBcb));
	KdPrint(("[EXT2]: Block bitmap\n"));
	DumpBlock(BlockGroup->BlockBitmap, Vcb->BLBlockSize, 1);
#endif
	// load the inode bitmap
#ifdef RWSUPPORT
	BlockGroup->INodeBitmap = XtBLGetBlocks(Vcb, BlockGroup->Info->bg_inode_bitmap, Vcb->BGINodeMapSize, &(BlockGroup->INodeBitmapBcb));
	KdPrint(("[EXT2]: Inode bitmap\n"));
	DumpBlock(BlockGroup->INodeBitmap, Vcb->BLBlockSize, 1);
#endif

	return STATUS_SUCCESS;
}

NTSTATUS
XtBGUnLoadGroup(
	PtrXtFSVCB Vcb,
	PBLOCK_GROUP BlockGroup,
	__u32 GroupNumber
)
{
#ifdef RWSUPPORT
	NYI
#endif
	return STATUS_SUCCESS;
}



/*
	if the group is not allocated then allocate it
	inc the ref count and set the Pointer
	return success

	group numbers are 1 based

	calls to this function need to be matched to a call to 
	XtBGFreeBlockGroup

*/
NTSTATUS 
XtBGGetBlockGroup(
	PtrXtFSVCB Vcb,
	__u32 GroupNumber,
	PBLOCK_GROUP *Pointer
)
{
	NTSTATUS Status;
	
	ASSERT(Vcb);
	ASSERT(Pointer);

	ACQUIRE_EX(Vcb->Resource);

	try
	{
		if(GroupNumber <= 0 ||  GroupNumber > Vcb->BGGroupCount)
		{
			KdPrint(("[EXT2]: BlockGroupNumber %d is out of range\n", GroupNumber));
			Status = STATUS_INVALID_PARAMETER;
		}

		*Pointer = Vcb->BGBlockGroups[GroupNumber - 1];
		if(! *Pointer)
		{
			// alloc the block group
			*Pointer = XtMMAllocatePool(NonPagedPool, sizeof(XtBGBLOCK_GROUP));
			if(*Pointer)
			{
				Vcb->BGBlockGroups[GroupNumber - 1] = (PBLOCK_GROUP)(*Pointer);
				XtBGLoadNewGroup(Vcb, *Pointer, GroupNumber);

				(*Pointer)->RefCount++;
				Status = STATUS_SUCCESS;
			}
			else
			{
				// out of memory?
				KdPrint(("[EXT2]: XtBGGetBlockGroup out of memory\n"));
				Status = STATUS_NO_MEMORY;
			}

		}
	}
	finally
	{
		RELEASE(Vcb->Resource);
	}

	return Status;
}

/*
	dec the ref count
	if ref count = 0 then deallocate it
	return success
*/
NTSTATUS 
XtBGFreeBlockGroup(
	PtrXtFSVCB Vcb,
	__u32 GroupNumber,
	PBLOCK_GROUP *Pointer
)
{
	ACQUIRE_EX(Vcb->Resource);

	try
	{
		if(*Pointer)
		{
			(*Pointer)->RefCount++;
			if((*Pointer)->RefCount == 0)
			{
				// free the group...
				Vcb->BGBlockGroups[GroupNumber - 1] = NULL;

				// release any resources...
				XtBGUnLoadGroup(Vcb, *Pointer, GroupNumber);
			
				XtMMFreePool(*Pointer, sizeof(XtBGBLOCK_GROUP));
				*Pointer = NULL;
			}

		}
	}
	finally
	{
		RELEASE(Vcb->Resource);
	}

	return STATUS_SUCCESS;
}


// does do not do any work, just calculates the group an
// inode can be found in

NTSTATUS
XtBGGetINodeInGroup(
	PtrXtFSVCB Vcb,
	__u32 INodeNumber,
	__u32 *GroupNumber,
	__u32 *INodeInGroup
)
{
	ASSERT(Vcb);
	ASSERT(GroupNumber);
	ASSERT(INodeInGroup);

	if(INodeNumber > 0 && INodeNumber <= Vcb->SuperBlock->s_inodes_count)
	{
		*GroupNumber = ((INodeNumber - 1) / Vcb->SuperBlock->s_inodes_per_group) + 1;
		*INodeInGroup = (INodeNumber - 1) % Vcb->SuperBlock->s_inodes_per_group;

		//KdPrint(("[EXT2]: %d == %d,%d\n", INodeNumber, *GroupNumber, *INodeInGroup));

		return STATUS_SUCCESS;
	}
	else
	{
		KdPrint(("[EXT2]: Inode %d is out of range (0, %d)\n", INodeNumber, Vcb->SuperBlock->s_inodes_count));
		return STATUS_INVALID_PARAMETER;
	}
}


// LockBlock
/*
	This is a general purpose block object which represents a 1k, 2k or 4k block on the disk
	This must be UnLocked in the same context as which it was Locked unless it is made 
	persistant with XtMakeBlockPersistant
*/
NTSTATUS
XtLockBlock(
	PtrXtFSVCB Vcb,
	__u32 BlockNumber,
	__u32 BlockCount,
	PBLOCK_OBJECT *Pointer
)
{
	NTSTATUS Result;
	void *start;
	PBLOCK_OBJECT Block;
	// look through the linked list...

	// Acquire the INode lock
	try
	{
		ACQUIRE_EX(Vcb->BlockResource);
		start = Vcb->NextBlockObject.Flink;

		while(start != &Vcb->NextBlockObject)
		{
			Block = CONTAINING_RECORD(start, XtBGBLOCK_OBJECT, NextObject);
			if(Block->BlockNo == BlockNumber && Block->BlockCount == BlockCount)
			{
				// we have a match
				*Pointer = Block;
				(*Pointer)->RefCount++;
				Result = STATUS_SUCCESS;
				try_return (STATUS_SUCCESS);
			}
			// advance to next record
			start = Block->NextObject.Flink;
		}


		// if not already found
		*Pointer = XtMMAllocatePoolStruct(NonPagedPool, sizeof(XtBGBLOCK_OBJECT), BLK);

		(*Pointer)->BlockNo = BlockNumber;
		(*Pointer)->BlockCount = BlockCount;
		(*Pointer)->RefCount++;

		(*Pointer)->Buffer = XtBLGetBlocks(Vcb, BlockNumber, BlockCount, &((*Pointer)->Bcb));

		// link into the list
		InsertTailList(&(Vcb->NextBlockObject), &((*Pointer)->NextObject));

		Result = STATUS_SUCCESS;

		try_exit:;
	}
	finally
	{
		RELEASE(Vcb->BlockResource);
	}

	// for testing I will make all blocks persistant...
	XtMakeBlockPersistant(Vcb, Pointer);
	// there must be something I have missed in the directory control code which causes a bsod

	return Result;
}

/*
	This copies the data into another buffer so that the original data can be unpinned
	This is because for some reason (good or bad) we can't unpin the data in a different
	context than it was pinned in
*/
NTSTATUS
XtMakeBlockPersistant(
	PtrXtFSVCB Vcb,
	PBLOCK_OBJECT *Pointer
)
{
	PVOID Buffer;
	ULONG Length;

	ASSERT(Pointer);
	if(*Pointer)
	{
		try
		{
			//KdPrint(("[EXT2]: XtMakeBlockPersistant\n"));
			ACQUIRE_EX(Vcb->BlockResource);

			if((*Pointer)->Bcb) // if there is no Bcb then this is done already
			{

				// allocate a buffer
				Length = (*Pointer)->BlockCount * Vcb->BLBlockSize; 
				Buffer = ExAllocatePool(NonPagedPool, Length);

				// Copy pinned data into buffer
				RtlCopyMemory(Buffer, (*Pointer)->Buffer, Length);

				// free the bcb based
				XtBLFreeBlocks(Vcb, &(*Pointer)->Bcb);

				// save the Buffer
				(*Pointer)->Buffer = Buffer;

				// we set the Bcb = NULL means this is not a Bcb
				(*Pointer)->Bcb = NULL;
			}

			//KdPrint(("[EXT2]: Done\n"));
		}
		finally
		{
			RELEASE(Vcb->BlockResource);
		}
	}
	return STATUS_SUCCESS;
}

// UnLockBlock
NTSTATUS
XtUnLockBlock(
	PtrXtFSVCB Vcb,
	PBLOCK_OBJECT *Pointer
)
{
	ASSERT(Pointer);

	if(*Pointer)
	{
		try
		{
			ACQUIRE_EX(Vcb->BlockResource);
			(*Pointer)->RefCount--;
			if((*Pointer)->RefCount == 0)
			{
				// Unlink from VCB
				RemoveEntryList(&(*Pointer)->NextObject);

				if((*Pointer)->Bcb)
				{
					//KdPrint(("[EXT2]: Freeing blocks\n"));
					XtBLFreeBlocks(Vcb, &(*Pointer)->Bcb);
				}
				else
				{
					ExFreePool((*Pointer)->Buffer);
				}
				//KdPrint(("[EXT2]: Freeing memory\n"));
				XtMMFreePool(*Pointer, sizeof(XtBGBLOCK_OBJECT));
				//KdPrint(("[EXT2]: Done\n"));
				*Pointer = NULL;
			}
		}
		finally
		{
			RELEASE(Vcb->BlockResource);
		}
	}
	return STATUS_SUCCESS;
}


// LockINode
NTSTATUS
XtLockINode(
	PtrXtFSVCB Vcb,
	__u32 INodeNumber,
	PINODE_OBJECT *Pointer
)
{
	NTSTATUS Result;
	__u32 GroupNumber;
	__u32 INodeInGroup;
	__u32 BlockNumber;
	__u32 INodeInBlock;
	__u32 INodeStartBlock;

	// look through the linked list...
	void *start;
	PINODE_OBJECT INode;

	// Acquire the INode lock
	try
	{
		ACQUIRE_EX(Vcb->INodeResource);

		start = Vcb->NextINodeObject.Flink;

		while(start != &Vcb->NextINodeObject)
		{
			INode = CONTAINING_RECORD(start, XtBGINODE_OBJECT, NextObject);
			if(INode->INodeNo == INodeNumber)
			{
				// we have a match
				// KdPrint(("[EXT2]: inode already locked\n"));
				*Pointer = INode;
				(*Pointer)->RefCount++;
				Result = STATUS_SUCCESS;
				try_return (Result);
			}
			// advance to next record
			start = INode->NextObject.Flink;
		}

		// if not already found
		*Pointer = XtMMAllocatePoolStruct(NonPagedPool, sizeof(XtBGINODE_OBJECT), INO);

		(*Pointer)->INodeNo = INodeNumber;
		(*Pointer)->RefCount++;

		// we need to locate the block in which this inode can be found...

		ASSERT(Vcb);
		ASSERT(Pointer);
		// KdPrint(("[EXT2]: locking inode\n"));

		// find out the inode & group number
		XtBGGetINodeInGroup(Vcb, INodeNumber, &GroupNumber, &INodeInGroup);

		// the inodes start at BlockGroup->Info.bg_inode_table
		// which is also in the Vcb descriptors
		// there are Vcb->INodesPerBlock INodes in each block
		INodeStartBlock = Vcb->BGGroupDescriptors[GroupNumber - 1].bg_inode_table;

		//KdPrint(("[EXT2]: bg_inode_table = %d\n", BlockGroup->Info->bg_inode_table));
		//KdPrint(("[EXT2]:                = %d\n", Vcb->BGGroupDescriptors[GroupNumber - 1].bg_inode_table));

		BlockNumber = INodeStartBlock + (INodeInGroup / Vcb->INodesPerBlock);
		INodeInBlock = INodeInGroup % Vcb->INodesPerBlock;
		//KdPrint(("[EXT2]: block = %d inode in block = %d\n", BlockNumber, INodeInBlock));

		// lock the required block
		XtLockBlock(Vcb, BlockNumber, 1, &((*Pointer)->Block));

		(*Pointer)->INode = ((PINODE)(*Pointer)->Block->Buffer) + INodeInBlock; // magic pointer addition

		// Link into the list
		InsertTailList(&(Vcb->NextINodeObject), &((*Pointer)->NextObject));

		Result = STATUS_SUCCESS;

		try_exit:;
	}
	finally
	{
		RELEASE(Vcb->INodeResource);
	}

	return Result;
}

NTSTATUS
XtMakeINodePersistant(
	PtrXtFSVCB Vcb,
	PINODE_OBJECT *Pointer
)
{
	ASSERT(Vcb);
	ASSERT(Pointer);

	if(*Pointer)
	{
		try
		{
			ACQUIRE_EX(Vcb->INodeResource);

			XtMakeBlockPersistant(Vcb, &((*Pointer)->Block));
		}
		finally
		{
			RELEASE(Vcb->INodeResource);
		}
	}
	return STATUS_SUCCESS;
}

// UnLockINode
NTSTATUS
XtUnLockINode(
	PtrXtFSVCB Vcb,
	PINODE_OBJECT *Pointer
)
{
	ASSERT(Vcb);
	ASSERT(Pointer);

	if(*Pointer)
	{
		try
		{
			ACQUIRE_EX(Vcb->INodeResource);

			(*Pointer)->RefCount--;
			if((*Pointer)->RefCount == 0)
			{
				// Unlink from VCB
				RemoveEntryList(&(*Pointer)->NextObject);

				XtUnLockBlock(Vcb, &((*Pointer)->Block));
				XtMMFreePool(*Pointer, sizeof(XtBGINODE_OBJECT));
				*Pointer = NULL;
			}
		}
		finally
		{
			RELEASE(Vcb->INodeResource);
		}
	}
	return STATUS_SUCCESS;
}

NTSTATUS
XtCreateDirObject(
	PtrXtFSVCB Vcb,
	PINODE_OBJECT INode,
	PDIR_OBJECT *Dir,
	PSTRING Filter
)
{
	ASSERT(INode);
	ASSERT(Dir);

	*Dir = XtMMAllocatePoolStruct(NonPagedPool, sizeof(XtBGDIR_OBJECT), DIR);

	(*Dir)->INode = INode;

	ACQUIRE_EX(Vcb->INodeResource);
	INode->RefCount++; // this means we need to call free...
	RELEASE(Vcb->INodeResource);

	// init enumeration state
	(*Dir)->BlockNumber = 0; // no block
	(*Dir)->Vcb = Vcb;
	(*Dir)->BlockSize = Vcb->BLBlockSize;
	(*Dir)->Offset = 0;

	// we have to make a local copy of the filter...(if there is one)
	if(Filter)
	{
		(*Dir)->Filter = ExAllocatePool(NonPagedPool, Filter->Length);
		RtlCopyMemory((*Dir)->Filter, Filter->Buffer, Filter->Length);
		(*Dir)->FilterLen = Filter->Length;
	}
	else
	{
		(*Dir)->Filter = NULL;
		(*Dir)->FilterLen = 0;
	}

	return STATUS_SUCCESS;
}

NTSTATUS
XtMakeDirObjectPersistant(
	PDIR_OBJECT *Dir
)
{
	ASSERT(Dir);

	if(*Dir)
	{
		if((*Dir)->Block)
		{
			XtMakeBlockPersistant((*Dir)->Vcb, &(*Dir)->Block);
		}
	}
	return STATUS_SUCCESS;
}

NTSTATUS
XtDestroyDirObject(
	PDIR_OBJECT *Dir
)
{
	ASSERT(Dir);

	if(*Dir)
	{
		XtUnLockINode((*Dir)->Vcb, &(*Dir)->INode);
		if((*Dir)->Filter)
		{
			ExFreePool((*Dir)->Filter);
		}
		XtMMFreePool(*Dir, sizeof(XtBGDIR_OBJECT));

		*Dir = 0;
	}
	return STATUS_SUCCESS;
}

// on entry, name_len must contain the length of name_buffer in bytes
// on exit it will contain the length of the string in name_buffer
NTSTATUS
XtDirGetEntry(
	PDIR_OBJECT Dir,
	WCHAR *name_buffer,
	ULONG *name_len,
	__u32 *inode
)
{
	char *data;
	struct ext2_dir_entry_2 *dir;
	ULONG j;
	int Index;
	NTSTATUS Status = STATUS_MORE_PROCESSING_REQUIRED;
	char str[512];
	int i;

	ASSERT(inode);
	ASSERT(name_buffer);
	ASSERT(Dir);

	//KdPrint(("[EXT2]: XtDirGetEntry\n"));
	// if there are no more entries then return STATUS_NO_MORE_FILES
	// else fill in info and return STATUS_SUCCESS

	while(Status == STATUS_MORE_PROCESSING_REQUIRED)
	{
		if((Dir->BlockNumber == 0) || (Dir->Offset >= Dir->BlockSize))
		{
			//KdPrint(("[EXT2]: Going to next block from %d\n", Dir->BlockNumber));
			// go to next block.
			if(Dir->Block)
			{
				//KdPrint(("[EXT2]: Unlocking previous block\n"));
				XtUnLockBlock(Dir->Vcb, &(Dir->Block));
			}
			Dir->BlockNumber++;

			if((Dir->BlockNumber > EXT2_NDIR_BLOCKS) || (Dir->INode->INode->i_block[Dir->BlockNumber - 1] == 0))
			{
				if(Dir->BlockNumber > EXT2_NDIR_BLOCKS)
				{
					KdPrint(("[EXT2]: indirect blocks NYI\n"));
				}
				Status = STATUS_NO_MORE_FILES;
				Dir->BlockNumber = 0;
				break;
			}
			//KdPrint(("[EXT2]: Locking block\n"));
			XtLockBlock(Dir->Vcb, Dir->INode->INode->i_block[Dir->BlockNumber - 1], 1, &(Dir->Block));

			Dir->Offset = 0;
			//KdPrint(("[EXT2]: ready to keep going\n"));
		}

		data = Dir->Block->Buffer;

		dir = (struct ext2_dir_entry_2 *)(data + Dir->Offset);

		if(dir->inode != 0)
		{
			Status = STATUS_SUCCESS;

			// check against the filter
			if(Dir->Filter)
			{
				/*i = 0;
				for(j = 0; j < Dir->FilterLen; j+=2)
				{
					str[i++] = Dir->Filter[j];
					//KdPrint(("[EXT2]: Filter test asdf %c\n", Dir->Filter[j]));
				}
				str[i++] = '-';
				for(j = 0; j < dir->name_len; j++) 
				{
					str[i++] = dir->name[j];
				}
				str[i++] = 0;
				if(i >= 512)
				{
					KdPrint(("[EXT2]: Opps, I overfilled the buffer!\n"));
				}
				KdPrint(("[EXT2]: Filter test %s\n", str));*/

				
				if(Dir->FilterLen == (ULONG)(dir->name_len * 2))
				{
					KdPrint(("[EXT2]: Checking against filter\n"));
					Status = STATUS_SUCCESS; // assume success until we find wrong data
					// check here
					for(Index = 0; Index < dir->name_len; Index++)
					{
						if(Dir->Filter[Index * 2] != dir->name[Index])
						{
							// names don't match
							//KdPrint(("[EXT2]: Filter does not match (wrong data %c != %c) (but len = %d)\n", Dir->Filter[Index * 2], dir->name[Index], Dir->FilterLen));
							Status = STATUS_MORE_PROCESSING_REQUIRED;
							break;
						}
					}
				}
				else
				{
					// names are not the same length
					Status = STATUS_MORE_PROCESSING_REQUIRED;
					//KdPrint(("[EXT2]: Filter does not match (wrong length)\n"));
				}
			}

			// testing...
/*			if(name_buffer[0] == '.' && dir->name_len == 1)
			{
				Status = STATUS_MORE_PROCESSING_REQUIRED;
			}*/
			// testing...

			// fill in the results
			if(Status == STATUS_SUCCESS) // there was a match with the filter
			{

				if(*name_len < (ULONG)(dir->name_len * 2))
				{
					KdPrint(("[EXT2]: file name buffer is too small %d < %d\n", *name_len, (dir->name_len * 2)));
					Status = STATUS_BUFFER_TOO_SMALL;
					break;
				}

				*inode = dir->inode;
				*name_len = (dir->name_len * 2);

				for(i = 0; i< dir->name_len; i++)
				{
					name_buffer[i] = dir->name[i];
					//KdPrint(("[EXT2]: Copy File Name %c\n", dir->name[i]));
				}
			}
		}

		// advance to next record
		Dir->Offset += dir->rec_len;
	}

	return Status;
}

/*
	This is a modified version of XtDirGetEntry
	which is called by XtDirFindEntry
*/
NTSTATUS
XtDirMatchEntry(
	PDIR_OBJECT Dir,
	PCHAR name_buffer,
	ULONG name_len,
	__u32 *inode
)
{
	char *data;
	struct ext2_dir_entry_2 *dir;
	int i;
	NTSTATUS Status = STATUS_MORE_PROCESSING_REQUIRED;

	ASSERT(inode);
	ASSERT(name_buffer);
	ASSERT(Dir);

//	KdPrint(("[EXT2]: XtDirGetEntry.  This needs to be checked\n"));
	// if there are no more entries then return STATUS_NO_MORE_FILES
	// else fill in info and return STATUS_SUCCESS

	while(Status == STATUS_MORE_PROCESSING_REQUIRED)
	{
		if((Dir->BlockNumber == 0) || (Dir->Offset >= Dir->BlockSize))
		{
			//KdPrint(("[EXT2]: Going to next block from %d\n", Dir->BlockNumber));
			// go to next block.
			if(Dir->Block)
			{
				//KdPrint(("[EXT2]: Unlocking previous block\n"));
				XtUnLockBlock(Dir->Vcb, &(Dir->Block));
			}
			Dir->BlockNumber++;

			if((Dir->BlockNumber > EXT2_NDIR_BLOCKS) || (Dir->INode->INode->i_block[Dir->BlockNumber - 1] == 0))
			{
				//KdPrint(("[EXT2]: no more blocks\n"));
				Status = STATUS_NO_MORE_FILES;
				break;
			}
			//KdPrint(("[EXT2]: Locking block again\n"));
			XtLockBlock(Dir->Vcb, Dir->INode->INode->i_block[Dir->BlockNumber - 1], 1, &(Dir->Block));

			Dir->Offset = 0;
			//KdPrint(("[EXT2]: ready to keep going\n"));
		}

		data = Dir->Block->Buffer;

		dir = (struct ext2_dir_entry_2 *)(data + Dir->Offset);

		if(dir->inode != 0)
		{
			// see if the file name matches
			*inode = dir->inode;
			//KdPrint(("[EXT2]: names %d %d\n", name_len, dir->name_len));

			if(name_len == dir->name_len)
			{
				if(RtlCompareMemory(dir->name, name_buffer, name_len) == name_len)
				{
					Status = STATUS_SUCCESS;
				}
			}
		}

		// advance to next record
		Dir->Offset += dir->rec_len;
	}
	return Status;
}
/*
	This will enumerate the directory and find an inode which match the supplied name
*/
NTSTATUS
XtDirFindEntry(
	PDIR_OBJECT Dir,
	PCHAR name_buffer,
	ULONG name_len,
	__u32 *inode
)
{
	NTSTATUS Status = STATUS_MORE_PROCESSING_REQUIRED;
	while(Status == STATUS_MORE_PROCESSING_REQUIRED)
	{
		Status = XtDirMatchEntry(Dir, name_buffer, name_len, inode);
	}
	return Status;
}

NTSTATUS
XtBLReadFile (
    IN PFSD_IRP_CONTEXT IrpContext,
	IN PtrXtFSVCB		Vcb,
	IN PtrXtFSCCB		Ccb,
    IN PLARGE_INTEGER   Offset,
    IN ULONG            Length,
    IN OUT PCHAR        Buffer
    )
{
	NTSTATUS Status = STATUS_NOT_IMPLEMENTED;
	__u32 ReadSoFar = 0;
	LARGE_INTEGER LargeSize;

	KdPrint(("[EXT2]: Reading %d bytes from %u\n", Length, Offset->QuadPart));

	// we should check that the inode is available and is a file
	ASSERT(Ccb);
	ASSERT(Ccb->fcb);
	ASSERT(Ccb->fcb->INode);
	ASSERT(Ccb->fcb->INode->INode);

	if(S_ISREG(Ccb->fcb->INode->INode->i_mode))
	{
		__u32 blockid;
		__u32 blockoffset;
		__u32 len;
		__u32 blockno;
		__u32 indblockno;
		__u32 dindblockno;
		PBLOCK_OBJECT Block;
		PBLOCK_OBJECT IBlock;
		PBLOCK_OBJECT DBlock;
		PBLOCK_OBJECT TBlock;
		ULONG Remaining;

		IrpContext->Irp->IoStatus.Status = STATUS_SUCCESS;
		IrpContext->Irp->IoStatus.Information = 1; // this seems to be what is done by other drivers
		Status = STATUS_SUCCESS;

		XtFSGetINodeSize(Ccb->fcb->vcb, Ccb->fcb->INode, &LargeSize);

		if(Offset->QuadPart + Length > LargeSize.QuadPart)
		{
			// we need to reduce the length...
			if(Offset->QuadPart >= LargeSize.QuadPart)
			{
				KdPrint(("[EXT2]: Read at or past end of file\n"));
				Length = 0;
				Status = STATUS_END_OF_FILE;
			}
			else
			{
				KdPrint(("[EXT2]: Read up to end of file\n"));
				Length = (ULONG)(LargeSize.QuadPart - Offset->QuadPart);
				Status = STATUS_SUCCESS; //STATUS_END_OF_FILE;
			}
		}



		Remaining = Length;

		while(Remaining > 0)
		{
			// see if the read is starting on a block boundry...
/*			if(((__u32)Offset->QuadPart + ReadSoFar) % Vcb->BLBlockSize)
			{
				KdPrint(("[EXT2]: Read does not start at block boundry (%d bytes in)\n", (((__u32)Offset->QuadPart + ReadSoFar) % Vcb->BLBlockSize)));
			}*/
		
			blockid = XtDiv64(Offset->QuadPart + ReadSoFar, Vcb->BLBlockSize, &blockoffset);
			KdPrint(("[EXT2]: Remaining = %d Want block %d offset %d\n", Remaining, blockid, blockoffset));

			if(blockid < EXT2_IND_BLOCK)
			{
				blockno = Ccb->fcb->INode->INode->i_block[blockid];
			}
			else if(blockid < Vcb->BLIndMax)
			{
				blockno = Ccb->fcb->INode->INode->i_block[EXT2_IND_BLOCK];
				//KdPrint(("[EXT2]: Reading the indirect block %d\n", blockno));
				// lock the indirect block into memory
				if(blockno == 0)
				{
					// must be sparse....?
					KdPrint(("[EXT2]: indirect block is 0\n"));
				}
				else
				{
					XtLockBlock(Vcb, blockno, 1, &IBlock);
					// read the appropriate block number out of it
					blockid -= EXT2_IND_BLOCK;
					//KdPrint(("[EXT2]: Want indirect block %d\n", blockid)); // we should sanity check this...
					blockno = ((__u32 *)IBlock->Buffer)[blockid];
	
					// unlock the indirect block
					XtUnLockBlock(Vcb, &IBlock);
				}
			}
			else if(blockid < Vcb->BLDIndMax)
			{
				blockno = Ccb->fcb->INode->INode->i_block[EXT2_DIND_BLOCK];
				if(blockno == 0)
				{
					KdPrint(("[EXT2]: dindirect block is 0\n"));
				}
				else
				{
					KdPrint(("[EXT2]: Reading the dindirect block %d\n", blockno));
					// lock the dindirect block into memory
					XtLockBlock(Vcb, blockno, 1, &DBlock);
					// read the appropriate block number out of it
					blockid -= Vcb->BLIndMax;

					// the dind block = blockid / N (where N = blocksize / 4)
					blockid = XtDiv1(blockid, Vcb->BLBlockSize / 4, &indblockno);
					KdPrint(("[EXT2]: Want indirect block %d, block number %d\n", blockid, indblockno)); // we should sanity check this...
					blockno = ((__u32 *)DBlock->Buffer)[blockid];
					KdPrint(("[EXT2]: blockno = %d\n", blockno));

					// unlock the dindirect block
					XtUnLockBlock(Vcb, &DBlock);

					if(blockno == 0)
					{
						KdPrint(("[EXT2]: dindirect->indirect block is 0\n"));
					}
					else
					{
						// lock the indirect block into memory
						XtLockBlock(Vcb, blockno, 1, &IBlock);
						blockno = ((__u32 *)IBlock->Buffer)[indblockno];
						KdPrint(("[EXT2]: blockno = %d\n", blockno));

						// unlock the indirect block
						XtUnLockBlock(Vcb, &IBlock);
					}
				}
			}
			else if(blockid < Vcb->BLTIndMax)
			{
				blockno = Ccb->fcb->INode->INode->i_block[EXT2_TIND_BLOCK];
				if(blockno == 0)
				{
					KdPrint(("[EXT2]: tindirect block is 0\n"));
				}
				else
				{
					KdPrint(("[EXT2]: Reading the tindirect block %d\n", blockno));
					// lock the tindirect block into memory
					XtLockBlock(Vcb, blockno, 1, &TBlock);
					blockid -= Vcb->BLDIndMax;

					blockid = XtDiv1(blockid, Vcb->BLBlockSize / 4, &indblockno);
					// now blockid contains ind & dind block and indblockno is ind index.
					blockid = XtDiv1(blockid, Vcb->BLBlockSize / 4, &dindblockno);
					// now blockid contains the dind block, dindblockno contains the dind index

					blockno = ((__u32 *)TBlock->Buffer)[blockid];

					// unlock the tindirect block
					XtUnLockBlock(Vcb, &TBlock);


					if(blockno == 0)
					{
						KdPrint(("[EXT2]: tindirect->dindirect block is 0\n"));
					}
					else
					{
						KdPrint(("[EXT2]: Reading the dindirect block %d\n", blockno));
						// lock the dindirect block into memory
						XtLockBlock(Vcb, blockno, 1, &DBlock);
		//DumpBlock(DBlock->Buffer, Vcb, 1);
		//Status = STATUS_END_OF_FILE;
						// read the appropriate block number out of it

						KdPrint(("[EXT2]: Want indirect block %d, block number %d\n", dindblockno, indblockno)); // we should sanity check this...
						blockno = ((__u32 *)DBlock->Buffer)[dindblockno];
						KdPrint(("[EXT2]: blockno = %d\n", blockno));
						
						// unlock the dindirect block
						XtUnLockBlock(Vcb, &DBlock);

						if(blockno == 0)
						{
							KdPrint(("[EXT2]: tindirect->dindirect->indirect block is 0\n"));
						}
						else
						{
							// lock the indirect block into memory
							XtLockBlock(Vcb, blockno, 1, &IBlock);
							blockno = ((__u32 *)IBlock->Buffer)[indblockno];
							KdPrint(("[EXT2]: blockno = %d\n", blockno));

							// unlock the indirect block
							XtUnLockBlock(Vcb, &IBlock);
						}
					}
				}
			}
			else
			{
				KdPrint(("[EXT2]: This block is too big.  Whats going on?\n"));
				Status = STATUS_END_OF_FILE;
			}

			if(Status == STATUS_SUCCESS)
			{
				if(Remaining > Vcb->BLBlockSize) // we only read one block at a time
				{
					len = Vcb->BLBlockSize;
				}
				else
				{
					len = Remaining;
				}

				if(len + blockoffset > Vcb->BLBlockSize)
				{
					// reduce the length because we are not reading the entire block
					len = Vcb->BLBlockSize - blockoffset;
				}

				//KdPrint(("[EXT2]: Reading %d bytes\n", len));
				// blockno now has the next block that we need to read...
				//KdPrint(("[EXT2]: Reading %d bytes of data from block %d\n", len, blockno));
				if(blockno == 0)
				{
					// this is a sparse file
					RtlFillMemory(Buffer, len, 0);
				}
				else
				{
					XtLockBlock(Vcb, blockno, 1, &Block);
					// copy the data

					RtlCopyMemory(Buffer, ((char *)Block->Buffer) + blockoffset, len);
					XtUnLockBlock(Vcb, &Block);
				}

				ReadSoFar += len;
				Remaining -= len;
				Buffer += len;
			}
			else
			{
				break;
			}


		}
		// update the file offset
		IrpContext->FileObject->CurrentByteOffset.QuadPart = Offset->QuadPart + Length;
		IrpContext->Irp->IoStatus.Information = Length;
		IrpContext->Irp->IoStatus.Status = STATUS_SUCCESS;
	}
	else
	{
		KdPrint(("[EXT2]: Not a regular file\n"));
		Status = STATUS_ACCESS_DENIED;
	}

	// is this a direct block....?

	// the trick here is that we can read at any part of the file so I need to map offsets into block
	// numbers...

	// complete the IRP

	XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);

	return Status;
}

// time functions
// nt times are 100ns counts since Jan 1 1601
// unix times are 1s counts since Jan 1 1970

#define OFFSET 0x019DB1DED53E8000L
#define FACTOR 10000000L

void XtUnixTimeToNTTime(__u32 *UnixTime, PLARGE_INTEGER NTTime)
{
	LARGE_INTEGER ut;
	ut.QuadPart = *UnixTime;
	NTTime->QuadPart = (ut.QuadPart * FACTOR) + OFFSET;
}

void XtNTTimeToUnixTime(PLARGE_INTEGER NTTime, __u32 *UnixTime)
{
	*UnixTime = (ULONG)(NTTime->QuadPart - OFFSET) / FACTOR;
}
