/*
	$Header: /home/itig/cvsroot/ext2ifs/src/fsctl.c,v 1.15 2003/01/27 05:31:30 jn Exp $
*/

#include "ext2fs.h"

// these are declerations local to this file

NTSTATUS
XtFSMountVolume (
    IN PFSD_IRP_CONTEXT IrpContext
    );

NTSTATUS
XtFSUserFsctl (
    IN PFSD_IRP_CONTEXT IrpContext
    );

NTSTATUS
XtFSVerifyVolume (
    IN PFSD_IRP_CONTEXT IrpContext
    );

//////////////////////

NTSTATUS 
FsdBlockDeviceIoControl (
    IN PDEVICE_OBJECT   DeviceObject,
    IN ULONG            IoctlCode,
    IN PVOID            InputBuffer,
    IN ULONG            InputBufferSize,
    IN OUT PVOID        OutputBuffer,
    IN OUT PULONG       OutputBufferSize
    )
{
    ULONG           OutputBufferSize2;
    KEVENT          Event;
    PIRP            Irp;
    IO_STATUS_BLOCK IoStatus;
    NTSTATUS        Status;

    PAGED_CODE();

    ASSERT(DeviceObject != NULL);

    if (OutputBufferSize)
    {
        OutputBufferSize2 = *OutputBufferSize;
    }

    KeInitializeEvent(&Event, NotificationEvent, FALSE);

    Irp = IoBuildDeviceIoControlRequest(
        IoctlCode,
        DeviceObject,
        InputBuffer,
        InputBufferSize,
        OutputBuffer,
        OutputBufferSize2,
        FALSE,
        &Event,
        &IoStatus
        );

    if (!Irp)
    {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    Status = IoCallDriver(DeviceObject, Irp);

    if (Status == STATUS_PENDING)
    {
        KeWaitForSingleObject(
            &Event,
            Suspended,
            KernelMode,
            FALSE,
            NULL
            );
        Status = IoStatus.Status;
    }

    if (OutputBufferSize)
    {
        *OutputBufferSize = IoStatus.Information;
    }

    return Status;
}




NTSTATUS
XtFSFileSystemControl (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
	NTSTATUS			Status;


	//PAGED_CODE();

    ASSERT(IrpContext);

    ASSERT((IrpContext->Identifier.Type == ICX) &&
           (IrpContext->Identifier.Size == sizeof(FSD_IRP_CONTEXT)));


	switch (IrpContext->MinorFunction)
	{
    case IRP_MN_USER_FS_REQUEST:
        Status = XtFSUserFsctl(IrpContext);
        break;

    case IRP_MN_MOUNT_VOLUME:
        Status = XtFSMountVolume(IrpContext);
        break;

    case IRP_MN_VERIFY_VOLUME:
        Status = XtFSVerifyVolume(IrpContext);
        break;

    default:
        KdPrint((DRIVER_NAME ": XtFSFileSystemControl: "
            "Unknown minor function 0x%x\n", IrpContext->MinorFunction));
        Status = STATUS_INVALID_DEVICE_REQUEST;//STATUS_NOT_IMPLEMENTED; // STATUS_INVALID_DEVICE_REQUEST ?
        IrpContext->Irp->IoStatus.Status = Status;
		XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);
    }

    return Status;
}


LARGE_INTEGER LargeIntegerZero = {0,0};
LARGE_INTEGER LargeIntegerMax = {MAXULONG,MAXLONG};


NTSTATUS
XtFSMountVolume (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
	NTSTATUS			Status;
	UNICODE_STRING		RegistryPath;
	PDEVICE_OBJECT		DeviceObject;
    PIRP                Irp;
    PIO_STACK_LOCATION  IoStack;
	PtrXtFSVCB			Vcb;
    CC_FILE_SIZES		FileSizes;
	ULONG				size;
	char				OrigStackSize;

	//PAGED_CODE();

    Irp = IrpContext->Irp;

	ASSERT(Irp);
        
    IoStack = IoGetCurrentIrpStackLocation(Irp);

	ASSERT(IoStack);

	DbgPrint("[EXT2]: FileSystemControl: IRP_MN_MOUNT_VOLUME\n");

	Status = XtSBCheckDeviceForExt2fs(
			IoStack->Parameters.MountVolume.DeviceObject
			);

	if(NT_SUCCESS(Status))
	{

		// read Drive Geometry (pg 609)

		// Create device object?
		Status = IoCreateDevice(
			Global->DriverObject,
			sizeof(XtFSVCB),
			0,
			FILE_DEVICE_DISK_FILE_SYSTEM,
			0,
			FALSE,
			&DeviceObject
			);

		if(NT_SUCCESS(Status))
		{
			DbgPrint("[EXT2]: Created device object for mounted FS\n");

			// init the VCB
			Vcb = (PtrXtFSVCB)DeviceObject->DeviceExtension;
			RtlZeroMemory(Vcb, sizeof(XtFSVCB));
			Vcb->Identifier.Type = VCB;
			ExInitializeResourceLite(&Vcb->Resource);

			Vcb->TargetDeviceObject = IoStack->Parameters.MountVolume.DeviceObject;
			Vcb->PtrVPB = IoStack->Parameters.MountVolume.Vpb;

			Vcb->PtrVPB->DeviceObject = DeviceObject;
			Vcb->PtrVPB->RealDevice = Vcb->TargetDeviceObject;
			Vcb->PtrVPB->Flags |= VPB_MOUNTED;

			DeviceObject->Vpb = Vcb->PtrVPB;
			DeviceObject->Flags &= (~DO_DEVICE_INITIALIZING);
			OrigStackSize = DeviceObject->StackSize;
			DeviceObject->StackSize = Vcb->TargetDeviceObject->StackSize + 1;

			ACQUIRE_EX(Global->Resource);
			InsertTailList(&(Global->NextVCB), &(Vcb->NextVCB));
			RELEASE(Global->Resource);

			Vcb->TargetDeviceObject->Flags &= (~DO_VERIFY_VOLUME);

			// create list for FCB's
			InitializeListHead(&Vcb->NextFCB);
			InitializeListHead(&Vcb->NextBlockObject);
			InitializeListHead(&Vcb->NextINodeObject);


			size = sizeof(DISK_GEOMETRY);

			FsdBlockDeviceIoControl(
				IoStack->Parameters.MountVolume.DeviceObject,
				IOCTL_DISK_GET_DRIVE_GEOMETRY,
				NULL,
				0,
				&Vcb->DiskGeometry,
				&size
				);

			KdPrint(("[EXT2]: DISK_GEOMETRY\n"));
			KdPrint(("[EXT2]: Cylinders         %d\n", Vcb->DiskGeometry.Cylinders));
			KdPrint(("[EXT2]: MediaType         %d\n", Vcb->DiskGeometry.MediaType));
			KdPrint(("[EXT2]: TracksPerCylinder %d\n", Vcb->DiskGeometry.TracksPerCylinder));
			KdPrint(("[EXT2]: SectorsPerTrack   %d\n", Vcb->DiskGeometry.SectorsPerTrack));
			KdPrint(("[EXT2]: BytesPerSector    %d\n", Vcb->DiskGeometry.BytesPerSector));

			if(Vcb->DiskGeometry.MediaType == FixedMedia)
			{
				//KdPrint(("[EXT2]: Fixed Media\n"));
			}
			else
			{
				KdPrint(("[EXT2]: Removable Media not supported\n"));
				Status = STATUS_UNRECOGNIZED_VOLUME;
			}

			if(NT_SUCCESS(Status))
			{
				size = sizeof(PARTITION_INFORMATION);

				FsdBlockDeviceIoControl(
					IoStack->Parameters.MountVolume.DeviceObject,
					IOCTL_DISK_GET_PARTITION_INFO,
					NULL,
					0,
					&Vcb->PartitionInformation,
					&size
					);

				KdPrint(("[EXT2]: PARTITION_INFORMATION\n"));
				KdPrint(("[EXT2]: StartingOffset      %d\n", Vcb->PartitionInformation.StartingOffset));
				KdPrint(("[EXT2]: PartitionLength     %u\n", Vcb->PartitionInformation.PartitionLength));
				KdPrint(("[EXT2]: HiddenSectors       %d\n", Vcb->PartitionInformation.HiddenSectors));
				KdPrint(("[EXT2]: PartitionNumber     %d\n", Vcb->PartitionInformation.PartitionNumber));
				KdPrint(("[EXT2]: PartitionType       %d\n", Vcb->PartitionInformation.PartitionType));
				KdPrint(("[EXT2]: BootIndicator       %d\n", Vcb->PartitionInformation.BootIndicator));
				KdPrint(("[EXT2]: RecognizedPartition %d\n", Vcb->PartitionInformation.RecognizedPartition));
				KdPrint(("[EXT2]: RewritePartition    %d\n", Vcb->PartitionInformation.RewritePartition));


				// set up the context for VirtualVolume cache
				Vcb->CommonFCBHeader.Flags = 0;
				Vcb->CommonFCBHeader.Flags2 = 0;
				Vcb->CommonFCBHeader.IsFastIoPossible = FastIoIsNotPossible;

				ExInitializeResource(&Vcb->MainResource);
				ExInitializeResource(&Vcb->PagingIoResource);

				Vcb->CommonFCBHeader.Resource = &(Vcb->MainResource);
				Vcb->CommonFCBHeader.PagingIoResource = &(Vcb->PagingIoResource);


				// init a file_object for the target device
				Vcb->VirtualVolumeFile = IoCreateStreamFileObject(NULL, Vcb->PtrVPB->RealDevice);
				ASSERT(Vcb->VirtualVolumeFile);

				Vcb->VirtualVolumeFile->Vpb = Vcb->PtrVPB;
				Vcb->VirtualVolumeFile->FsContext  = &(Vcb->CommonFCBHeader);
				Vcb->VirtualVolumeFile->FsContext2 = NULL;

				Vcb->VirtualVolumeFile->SectionObjectPointer = &Vcb->SectionObjectPointers;

				Vcb->VirtualVolumeFile->PrivateCacheMap = NULL;


				Vcb->VirtualVolumeFile->ReadAccess   = TRUE;
				Vcb->VirtualVolumeFile->WriteAccess  = FALSE;
				Vcb->VirtualVolumeFile->DeleteAccess = FALSE;

				Vcb->CommonFCBHeader.AllocationSize.QuadPart  = //1024;//sizeof(SUPER_BLOCK) + 1024;
				Vcb->CommonFCBHeader.FileSize.QuadPart        = //sizeof(SUPER_BLOCK) + 1024;
				Vcb->CommonFCBHeader.ValidDataLength.QuadPart = Vcb->PartitionInformation.PartitionLength.QuadPart;//LargeIntegerMax;

				FileSizes.AllocationSize  = Vcb->CommonFCBHeader.AllocationSize;
				FileSizes.FileSize        = Vcb->CommonFCBHeader.FileSize;
				FileSizes.ValidDataLength = Vcb->CommonFCBHeader.ValidDataLength;


				DbgPrint("[EXT2]: Initializing Cache Map\n");
				CcInitializeCacheMap( Vcb->VirtualVolumeFile,
									  &FileSizes,
									  TRUE,
									  &Global->CacheManagerNoOpCallbacks,
									  Vcb);

				// check the PrivateCacheMap to see if it worked...
				ASSERT(Vcb->VirtualVolumeFile->PrivateCacheMap);

				Status = XtSBMountFS(Vcb);

			}


			if(NT_SUCCESS(Status))
			{
				Vcb->TargetDeviceObject->Flags |= DO_VERIFY_VOLUME;
				DbgPrint("[EXT2]: Mount complete,\n");
			}
			else
			{
				DbgPrint("[EXT2]: Mount failed\n");
				// undo all the hard work...
				Vcb->PtrVPB->Flags &= (~VPB_MOUNTED);


//			Vcb->PtrVPB->DeviceObject = 0;
//			Vcb->PtrVPB->RealDevice = 0;

			DeviceObject->Vpb = 0;
			//DeviceObject->Flags &= (~DO_DEVICE_INITIALIZING);
//			DeviceObject->StackSize = Vcb->TargetDeviceObject->StackSize;
			DeviceObject->StackSize = OrigStackSize;



				ACQUIRE_EX(Global->Resource);
				RemoveEntryList(&(Vcb->NextVCB));
				RELEASE(Global->Resource);
				ExDeleteResourceLite(&Vcb->Resource);

			RtlZeroMemory(Vcb, sizeof(XtFSVCB));

				IoDeleteDevice(DeviceObject);

			}

		}
		else
		{
			DbgPrint("[EXT2]: Mount error\n");
		}
	}

    if (!NT_SUCCESS(Status))
    {
        Status = STATUS_UNRECOGNIZED_VOLUME;
    }
            
	Irp->IoStatus.Status = Status;
	Irp->IoStatus.Information = 0;
	XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);

	return Status;
}

#ifdef DO_REPARSE

#define IO_REPARSE_TAG_MOUNT_POINT              (0xA0000003L)       

NTSTATUS
XtFSUserFsctlGetReparsePoint (
	IN PFSD_IRP_CONTEXT IrpContext
	)
{
	NTSTATUS Status = STATUS_NOT_IMPLEMENTED;

    PEXTENDED_IO_STACK_LOCATION IrpSp = (PEXTENDED_IO_STACK_LOCATION)IoGetCurrentIrpStackLocation( IrpContext->Irp );
	int BufferSize;
	PREPARSE_DATA_BUFFER OutputBuffer;

	KdPrint(("[EXT2]: Getting reparse point info...\n"));
	KdPrint(("[EXT2]: In buffer size = %d\n", IrpSp->Parameters.FileSystemControl.InputBufferLength));
	KdPrint(("[EXT2]: Out buffer size = %d\n", IrpSp->Parameters.FileSystemControl.OutputBufferLength));

	BufferSize = IrpSp->Parameters.FileSystemControl.OutputBufferLength;

	// the in buffer is SystemBuffer
	// the out buffer is UserBuffer
	OutputBuffer = XtFSGetUserBuffer(IrpContext->Irp);
	RtlFillMemory(OutputBuffer, BufferSize, 0);

	if(BufferSize < sizeof(REPARSE_DATA_BUFFER))
	{
		KdPrint(("[EXT2]: Out buffer size is too small\n"));
		Status = STATUS_BUFFER_OVERFLOW;
	}
	else
	{
		OutputBuffer->ReparseTag = IO_REPARSE_TAG_MOUNT_POINT;
		OutputBuffer->ReparseDataLength = sizeof(REPARSE_DATA_BUFFER) + OutputBuffer->MountPointReparseBuffer.PrintNameLength;
		OutputBuffer->MountPointReparseBuffer.PrintNameLength = 1;
		OutputBuffer->MountPointReparseBuffer.PathBuffer[0] = 'J';
		OutputBuffer->MountPointReparseBuffer.PrintNameOffset = (char *)OutputBuffer->MountPointReparseBuffer.PathBuffer - (char *)OutputBuffer;

		Status = STATUS_SUCCESS;
	}


	return Status;
}

#endif

NTSTATUS
XtFSUserFsctl (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
	NTSTATUS Status = STATUS_NOT_IMPLEMENTED;

    PIO_STACK_LOCATION IrpSp = IoGetCurrentIrpStackLocation( IrpContext->Irp );

	KdPrint(("[EXT2]: XtFSUserFsctl request 0x%08x\n", ((PEXTENDED_IO_STACK_LOCATION)IrpSp)->Parameters.FileSystemControl.FsControlCode));

	switch(((PEXTENDED_IO_STACK_LOCATION)IrpSp)->Parameters.FileSystemControl.FsControlCode)
	{
	case FSCTL_LOCK_VOLUME:
		KdPrint(("[EXT2]: FSCTL_LOCK_VOLUME\n"));
		Status = STATUS_SUCCESS;
		break;

	case FSCTL_UNLOCK_VOLUME:
		KdPrint(("[EXT2]: FSCTL_UNLOCK_VOLUME\n"));
		Status = STATUS_SUCCESS;
		break;

	case FSCTL_DISMOUNT_VOLUME:
		KdPrint(("[EXT2]: FSCTL_DISMOUNT_VOLUME\n"));
		break;

	case FSCTL_MARK_VOLUME_DIRTY:
		KdPrint(("[EXT2]: FSCTL_MARK_VOLUME_DIRTY\n"));
		Status = STATUS_SUCCESS;
		break;

	case FSCTL_IS_VOLUME_MOUNTED:
		KdPrint(("[EXT2]: FSCTL_IS_VOLUME_MOUNTED\n"));
		Status = STATUS_SUCCESS;
		break;

	case FSCTL_IS_PATHNAME_VALID:
		KdPrint(("[EXT2]: FSCTL_IS_PATHNAME_VALID\n"));
		Status = STATUS_SUCCESS;
		break;

	case FSCTL_GET_RETRIEVAL_POINTERS:
		KdPrint(("[EXT2]: FSCTL_GET_RETRIEVAL_POINTERS\n"));
		Status = STATUS_INVALID_PARAMETER;
		break;

#ifdef DO_REPARSE
	case FSCTL_GET_REPARSE_POINT:
		KdPrint(("[EXT2]: FSCTL_GET_REPARSE_POINT\n"));
		Status = XtFSUserFsctlGetReparsePoint(IrpContext);

		break;
#endif

	default:
		KdPrint(("[EXT2]: Unknown FSCTL %d\n", (((PEXTENDED_IO_STACK_LOCATION)IrpSp)->Parameters.FileSystemControl.FsControlCode >> 2) & 0xFFF));
		Status = STATUS_INVALID_DEVICE_REQUEST;
	}

	IrpContext->Irp->IoStatus.Status = Status;
	IrpContext->Irp->IoStatus.Information = 0;
	XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);

	return Status;
}

NTSTATUS
XtFSVerifyVolume (
    IN PFSD_IRP_CONTEXT IrpContext
    )
{
	NTSTATUS Status = STATUS_SUCCESS;//INVALID_DEVICE_REQUEST;//STATUS_NOT_IMPLEMENTED;

	//PAGED_CODE();

	DbgPrint("[EXT2]: XtFSVerifyVolume\n");

	IrpContext->Irp->IoStatus.Status = Status;
	IrpContext->Irp->IoStatus.Information = 0;
	XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);

	return Status;
}
