/*
	$Header: /home/itig/cvsroot/ext2ifs/src/volinfo.c,v 1.9 2003/01/06 08:08:55 jn Exp $
*/

#include "ext2fs.h"

NTSTATUS
XtQueryFsSizeInfo (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PVCB Vcb,
    IN PFILE_FS_SIZE_INFORMATION Buffer,
    IN OUT PULONG Length
    )
{
    Buffer->TotalAllocationUnits.QuadPart = Vcb->SuperBlock->s_blocks_count;
    Buffer->AvailableAllocationUnits.QuadPart = Vcb->SuperBlock->s_free_blocks_count;
    Buffer->SectorsPerAllocationUnit = Vcb->BLBlockSize / Vcb->DiskGeometry.BytesPerSector;
    Buffer->BytesPerSector = Vcb->DiskGeometry.BytesPerSector;

    *Length -= sizeof( FILE_FS_SIZE_INFORMATION );

    return STATUS_SUCCESS;
}

NTSTATUS
XtQueryFsVolumeInfo (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PVCB Vcb,
    IN PFILE_FS_VOLUME_INFORMATION Buffer,
    IN OUT PULONG Length
    )
{
    NTSTATUS Status = STATUS_SUCCESS;
	int volume_name_len;
	int i;

	Buffer->VolumeCreationTime.QuadPart = 0;
	Buffer->VolumeSerialNumber = 0;
	Buffer->SupportsObjects = FALSE;
	//Buffer->VolumeLabelLength = 8;


	volume_name_len = 0;
	while(volume_name_len < 16)
	{
		if(Vcb->SuperBlock->s_volume_name[volume_name_len])
		{
			volume_name_len++;
		}
		else break;
	}
	Buffer->VolumeLabelLength = volume_name_len * 2;

	KdPrint(("[ETX2]: XtQueryFsVolumeInfo...\n"));
    *Length -= (sizeof( FILE_FS_VOLUME_INFORMATION ) - sizeof(WCHAR));
	KdPrint(("[ETX2]: XtQueryFsVolumeInfo.  Checking buffer size\n"));

	if((*Length) >= Buffer->VolumeLabelLength)
	{
		KdPrint(("[ETX2]: Enough space for volume label\n"));

		/*Buffer->VolumeLabel[0] = 'j';
		Buffer->VolumeLabel[1] = 'o';
		Buffer->VolumeLabel[2] = 'h';
		Buffer->VolumeLabel[3] = 'n';*/
		for(i = 0; i < volume_name_len; i++)
		{
			Buffer->VolumeLabel[i] = Vcb->SuperBlock->s_volume_name[i];
		}

		*Length -= Buffer->VolumeLabelLength;
	}
	else
	{
		KdPrint(("[ETX2]: Not enough space for volume label\n"));
		KdPrint(("[ETX2]: Not enough space for volume label (len = %d)\n", (*Length)));

		Buffer->VolumeLabelLength = 0;
		Status = STATUS_BUFFER_OVERFLOW;
	}

	return Status;
}

NTSTATUS
XtQueryFsDeviceInfo (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PVCB Vcb,
    IN PFILE_FS_DEVICE_INFORMATION Buffer,
    IN OUT PULONG Length
    )
{
    Buffer->Characteristics = Vcb->TargetDeviceObject->Characteristics;
    Buffer->DeviceType = FILE_DEVICE_DISK;

    *Length -= sizeof( FILE_FS_DEVICE_INFORMATION );

    return STATUS_SUCCESS;
}

#define EXT2_NAME L"EXT2"
#define EXT3_NAME L"EXT3"

NTSTATUS
XtQueryFsAttributeInfo (
    IN PFSD_IRP_CONTEXT IrpContext,
    IN PVCB Vcb,
    IN PFILE_FS_ATTRIBUTE_INFORMATION Buffer,
    IN OUT PULONG Length
    )
{
    NTSTATUS Status = STATUS_SUCCESS;
	WCHAR *DriverName = EXT2_NAME; // len = 8
    ULONG BytesToCopy;

	Buffer->FileSystemAttributes = FILE_CASE_SENSITIVE_SEARCH | FILE_CASE_PRESERVED_NAMES;
#ifdef DO_REPARSE
	Buffer->FileSystemAttributes |= FILE_SUPPORTS_REPARSE_POINTS;
#endif
	Buffer->MaximumComponentNameLength = EXT2_NAME_LEN;

	*Length -= (sizeof(FILE_FS_ATTRIBUTE_INFORMATION) - sizeof(WCHAR));

	if(*Length >= 8)
	{
		// Enough room for the name
		BytesToCopy = 8;
	}
	else
	{
		BytesToCopy = *Length;
		Status = STATUS_BUFFER_OVERFLOW;
	}

	Buffer->FileSystemNameLength = BytesToCopy;

	if(Vcb->SuperBlock->s_rev_level > 0)
	{
		if(Vcb->SuperBlock->s_feature_compat && EXT3_FEATURE_COMPAT_HAS_JOURNAL)
		{
			DriverName = EXT3_NAME; // len = 8
		}
	}

	RtlCopyMemory(
			Buffer->FileSystemName,
			DriverName,
			BytesToCopy	);

	*Length -= BytesToCopy;

	return Status;
}

NTSTATUS
XtFSQueryVolumeInformation (
    IN PFSD_IRP_CONTEXT IrpContext
	)
{
    IN PIRP             Irp = IrpContext->Irp;
    PIO_STACK_LOCATION  IoStack;
	PVOID				Buffer;
	ULONG				BufferLen;
	PFCB				Fcb;
	PCCB				Ccb;
	PVCB				Vcb;

	NTSTATUS			Status = STATUS_INVALID_PARAMETER;


	ASSERT(Irp);
        
    IoStack = IoGetCurrentIrpStackLocation(Irp);

	ASSERT(IoStack);

	// see if this is the VCB
	if(IoStack->FileObject->FsContext2 == NULL)
	{
		KdPrint(("[ETX2]: XtFSQueryVolumeInformationon on VCB\n"));
		Vcb = (PtrXtFSVCB)IrpContext->DeviceObject->DeviceExtension;
		ASSERT(Vcb);
		VERIFY(Vcb, VCB);
	}
	else
	{
		Ccb = (PtrXtFSCCB)IoStack->FileObject->FsContext2;
		Fcb = Ccb->fcb;
		Vcb = Fcb->vcb;

		ASSERT(Vcb);
		VERIFY(Vcb, VCB);
	}

	Buffer = Irp->AssociatedIrp.SystemBuffer;
	BufferLen = IoStack->Parameters.QueryVolume.Length;

	RtlZeroMemory(Buffer, BufferLen);

	switch(IoStack->Parameters.QueryVolume.FsInformationClass)
	{
		case FileFsVolumeInformation:
			KdPrint(("[ETX2]: FileFsVolumeInformation\n"));
			Status = XtQueryFsVolumeInfo( IrpContext, Vcb, Buffer, &BufferLen );
			break;

		case FileFsSizeInformation:
			KdPrint(("[ETX2]: FileFsSizeInformation\n"));
	        Status = XtQueryFsSizeInfo( IrpContext, Vcb, Buffer, &BufferLen );
			break;

		case FileFsDeviceInformation:
			KdPrint(("[ETX2]: FileFsDeviceInformation\n"));
	        Status = XtQueryFsDeviceInfo( IrpContext, Vcb, Buffer, &BufferLen );
			break;

		case FileFsAttributeInformation:
			KdPrint(("[ETX2]: FileFsAttributeInformation\n"));
	        Status = XtQueryFsAttributeInfo( IrpContext, Vcb, Buffer, &BufferLen );
			break;

		default:
			KdPrint((DRIVER_NAME ": FsdQueryVolumeInformation: "
				"Unknown FsInformationClass %d\n", IoStack->Parameters.QueryVolume.FsInformationClass));

			Status = STATUS_NOT_IMPLEMENTED;

	}
	Irp->IoStatus.Information = IoStack->Parameters.QueryVolume.Length - BufferLen;
	

	Irp->IoStatus.Status = Status;
	XtFSCompleteRequest(IrpContext, IO_NO_INCREMENT);

	return Status;
}
